#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six


def _python_2_unicode_compatible(class_):
    """
    A decorator that defines __unicode__ and __str__ methods under Python 2.
    Under Python 3 it does nothing.

    To support Python 2 and 3 with a single code base, define a __str__ method
    returning text and apply this decorator to the class.

    Copyright (c) 2010-2015 Benjamin Peterson
    """
    if six.PY2:
        if '__str__' not in class_.__dict__:
            raise ValueError("@python_2_unicode_compatible cannot be applied "
                             "to %s because it doesn't define __str__()." %
                             class_.__name__)
        class_.__unicode__ = class_.__str__
        class_.__str__ = lambda self: self.__unicode__().encode('utf-8')
    return class_

if not hasattr(six, 'python_2_unicode_compatible'):
    six.python_2_unicode_compatible = _python_2_unicode_compatible



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import specs


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('*equal')
def left_eq_null(left, right):
    return False


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('#operator_<')
def left_lt_null(left, right):
    return False


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('#operator_<=')
def left_lte_null(left, right):
    return False


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('#operator_>')
def left_gt_null(left, right):
    return True


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('#operator_>=')
def left_gte_null(left, right):
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('*equal')
def null_eq_right(left, right):
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('#operator_<')
def null_lt_right(left, right):
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('#operator_<=')
def null_lte_right(left, right):
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('#operator_>')
def null_gt_right(left, right):
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('#operator_>=')
def null_gte_right(left, right):
    return False


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('*not_equal')
def left_neq_null(left, right):
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('*not_equal')
def null_neq_right(left, right):
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('*equal')
def null_eq_null(left, right):
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('*not_equal')
def null_neq_null(left, right):
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('#operator_<')
def null_lt_null(left, right):
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('#operator_<=')
def null_lte_null(left, right):
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('#operator_>')
def null_gt_null(left, right):
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('#operator_>=')
def null_gte_null(left, right):
    return True


def register(context):
    context.register_function(left_eq_null)
    context.register_function(left_neq_null)
    context.register_function(left_lt_null)
    context.register_function(left_lte_null)
    context.register_function(left_gt_null)
    context.register_function(left_gte_null)

    context.register_function(null_eq_right)
    context.register_function(null_neq_right)
    context.register_function(null_lt_right)
    context.register_function(null_lte_right)
    context.register_function(null_gt_right)
    context.register_function(null_gte_right)

    context.register_function(null_eq_null)
    context.register_function(null_neq_null)
    context.register_function(null_lt_null)
    context.register_function(null_lte_null)
    context.register_function(null_gt_null)
    context.register_function(null_gte_null)



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from yaql.language import exceptions
from yaql.language import runner
from yaql.language import specs
from yaql.language import utils


class ContextBase(object):
    def __init__(self, parent_context):
        self._parent_context = parent_context

    @property
    def parent(self):
        return self._parent_context

    def register_function(self, spec, *args, **kwargs):
        pass

    def __getitem__(self, name):
        return None

    def __setitem__(self, name, value):
        pass

    def __delitem__(self, name):
        pass

    def __call__(self, name, engine, sender=utils.NO_VALUE,
                 data_context=None, return_context=False,
                 use_convention=False):
        raise exceptions.NoFunctionRegisteredException(name)

    def get_functions(self, name, predicate=None, use_convention=False):
        return []

    def collect_functions(self, name, predicate=None, use_convention=False):
        return [[]]

    def create_child_context(self):
        return type(self)(self)


class Context(ContextBase):
    def __init__(self, parent_context=None, data=utils.NO_VALUE,
                 convention=None):
        super(Context, self).__init__(parent_context)
        self._functions = {}
        self._data = {}
        self._convention = convention
        if data is not utils.NO_VALUE:
            self['$'] = data

    @staticmethod
    def _import_function_definition(fd):
        return fd

    def register_function(self, spec, *args, **kwargs):
        exclusive = kwargs.pop('exclusive', False)

        if not isinstance(spec, specs.FunctionDefinition) \
                and six.callable(spec):
            spec = specs.get_function_definition(
                spec, *args, convention=self._convention, **kwargs)

        spec = self._import_function_definition(spec)
        if spec.is_method:
            if not spec.is_valid_method():
                raise exceptions.InvalidMethodException(spec.name)
        self._functions.setdefault(spec.name, list()).append((spec, exclusive))

    def get_functions(self, name, predicate=None, use_convention=False):
        if use_convention and self._convention is not None:
            name = self._convention.convert_function_name(name)
        if predicate is None:
            predicate = lambda x: True
        return six.moves.filter(predicate, list(six.moves.map(
            lambda x: x[0],
            self._functions.get(name, list()))))

    def collect_functions(self, name, predicate=None, use_convention=False):
        if use_convention and self._convention is not None:
            name = self._convention.convert_function_name(name)
        overloads = []
        p = self
        while p is not None:
            layer_overloads = p._functions.get(name)
            p = p.parent
            if layer_overloads:
                layer = []
                for spec, exclusive in layer_overloads:
                    if exclusive:
                        p = None
                    if predicate and not predicate(spec):
                        continue
                    layer.append(spec)
                if layer:
                    overloads.append(layer)
        return overloads

    def __call__(self, name, engine, sender=utils.NO_VALUE,
                 data_context=None, return_context=False,
                 use_convention=False):
        return lambda *args, **kwargs: runner.call(
            name, self, args, kwargs, engine, sender,
            data_context, return_context, use_convention)

    @staticmethod
    def _normalize_name(name):
        if not name.startswith('$'):
            name = ('$' + name)
        if name == '$':
            name = '$1'
        return name

    def __setitem__(self, name, value):
        self._data[self._normalize_name(name)] = value

    def __getitem__(self, name):
        name = self._normalize_name(name)
        if name in self._data:
            return self._data[name]
        if self.parent:
            return self.parent[name]

    def __delitem__(self, name):
        self._data.pop(self._normalize_name(name))

    def create_child_context(self):
        return Context(self, convention=self._convention)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

import six

from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes


@specs.parameter('name', yaqltypes.StringConstant())
@specs.name('#get_context_data')
def get_context_data(name, context):
    return context[name]


@specs.parameter('sender', yaqltypes.Lambda(with_context=True,
                                            return_context=True))
@specs.parameter('expr', yaqltypes.Lambda(with_context=True, method=True,
                                          return_context=True))
@specs.name('#operator_.')
@specs.returns_context
def op_dot(context, sender, expr):
    return expr(*sender(context))


@specs.parameter('sender',
                 yaqltypes.Lambda(with_context=True, return_context=True))
@specs.parameter('expr', yaqltypes.YaqlExpression())
@specs.inject('operator', yaqltypes.Delegate('#operator_.', with_context=True))
@specs.name('#operator_?.')
def elvis_operator(context, operator, sender, expr):
    sender, context = sender(context)
    if sender is None:
        return None
    return operator(context, sender, expr)


@specs.parameter('sequence', yaqltypes.Iterable())
@specs.parameter('args', yaqltypes.String())
@specs.method
def unpack(sequence, context, *args):
    lst = tuple(itertools.islice(sequence, len(args) + 1))
    if 0 < len(args) != len(lst):
        raise ValueError('Cannot unpack {0} elements into {1}'.format(
            len(lst), len(args)))
    if len(args) > 0:
        for i in range(len(lst)):
            context[args[i]] = lst[i]
    else:
        for i, t in enumerate(sequence, 1):
            context[str(i)] = t
    return lst


def with_(context, *args):
    for i, t in enumerate(args, 1):
        context[str(i)] = t


@specs.inject('__context__', yaqltypes.Context())
def let(__context__, *args, **kwargs):
    for i, value in enumerate(args, 1):
        __context__[str(i)] = value

    for key, value in six.iteritems(kwargs):
        __context__[key] = value


@specs.parameter('name', yaqltypes.String())
@specs.parameter('func', yaqltypes.Lambda())
def def_(name, func, context):
    @specs.name(name)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    context.register_function(wrapper)


@specs.parameter('left', yaqltypes.Lambda(return_context=True))
@specs.parameter('right', yaqltypes.Lambda(with_context=True))
@specs.name('#operator_->')
def send_context(left, right):
    context = left()[1]
    return right(context)


@specs.method
@specs.parameter('condition', yaqltypes.Lambda())
@specs.parameter('message', yaqltypes.String())
def assert_(engine, obj, condition, message=u'Assertion failed'):
    if utils.is_iterator(obj):
        obj = utils.memorize(obj, engine)
    if not condition(obj):
        raise AssertionError(message)
    return obj


def register(context):
    context.register_function(get_context_data)
    context.register_function(op_dot)
    context.register_function(unpack)
    context.register_function(with_)
    context.register_function(send_context)
    context.register_function(let)
    context.register_function(def_)
    context.register_function(elvis_operator)
    context.register_function(assert_)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaql
from yaql.language import factory
from yaql.standard_library import legacy as std_legacy


class YaqlFactory(factory.YaqlFactory):
    def __init__(self):
        # noinspection PyTypeChecker
        super(YaqlFactory, self).__init__(keyword_operator=None)
        self.insert_operator(
            'or', True, '=>',
            factory.OperatorType.BINARY_LEFT_ASSOCIATIVE, True)

    def create(self, options=None):
        options = dict(options or {})
        options['yaql.convertTuplesToLists'] = False
        return super(YaqlFactory, self).create(options)


def create_context(*args, **kwargs):
    tuples = kwargs.pop('tuples', True)

    context = yaql.create_context(*args, **kwargs)
    context = context.create_child_context()

    std_legacy.register(context, tuples)
    return context



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes
import yaql.standard_library.queries


@specs.parameter('args', nullable=True)
@specs.inject('delegate', yaqltypes.Delegate('to_list', method=True))
def list_(delegate, *args):
    def rec(seq):
        for t in seq:
            if utils.is_iterator(t):
                for t2 in rec(t):
                    yield t2
            else:
                yield t
    return delegate(rec(args))


@specs.method
@specs.parameter('collection', yaqltypes.Iterable())
def to_list(collection):
    if isinstance(collection, tuple):
        return collection
    return tuple(collection)


@specs.parameter('args', nullable=True)
@specs.name('#list')
def build_list(engine, *args):
    utils.limit_memory_usage(engine, *((1, t) for t in args))
    return tuple(args)


@specs.no_kwargs
@specs.parameter('args', utils.MappingRule)
def dict_(engine, *args):
    result = {}
    for t in args:
        result[t.source] = t.destination
        utils.limit_memory_usage(engine, (1, result))
    return utils.FrozenDict(result)


@specs.parameter('items', yaqltypes.Iterable())
@specs.no_kwargs
def dict__(items, engine):
    result = {}
    for t in items:
        it = iter(t)
        key = next(it)
        value = next(it)
        result[key] = value
        utils.limit_memory_usage(engine, (1, result))
    return utils.FrozenDict(result)


@specs.parameter('collection', yaqltypes.Iterable())
@specs.parameter('key_selector', yaqltypes.Lambda())
@specs.parameter('value_selector', yaqltypes.Lambda())
@specs.method
def to_dict(collection, engine, key_selector, value_selector=None):
    result = {}
    for t in collection:
        key = key_selector(t)
        value = t if value_selector is None else value_selector(t)
        result[key] = value
        utils.limit_memory_usage(engine, (1, result))
    return result


@specs.parameter('d', utils.MappingType, alias='dict', nullable=True)
@specs.parameter('key', yaqltypes.Keyword())
@specs.name('#operator_.')
def dict_keyword_access(d, key):
    return d[key]


@specs.parameter('collection', yaqltypes.Sequence())
@specs.parameter('attribute', yaqltypes.Keyword(expand=False))
@specs.inject('operator', yaqltypes.Delegate('#operator_.'))
@specs.name('#operator_.')
def collection_attribute(collection, attribute, operator):
    return six.moves.map(
        lambda t: operator(t, attribute), collection)


@specs.parameter('d', utils.MappingType, alias='dict')
@specs.name('#indexer')
def dict_indexer(d, key):
    return d[key]


@specs.parameter('d', utils.MappingType, alias='dict')
@specs.name('#indexer')
def dict_indexer_with_default(d, key, default):
    return d.get(key, default)


@specs.parameter('d', utils.MappingType, alias='dict')
@specs.name('get')
@specs.method
def dict_get(d, key, default=None):
    return d.get(key, default)


@specs.parameter('d', utils.MappingType, alias='dict')
@specs.name('keys')
@specs.method
def dict_keys(d):
    return six.iterkeys(d)


@specs.parameter('d', utils.MappingType, alias='dict')
@specs.name('values')
@specs.method
def dict_values(d):
    return six.itervalues(d)


@specs.parameter('d', utils.MappingType, alias='dict')
@specs.name('items')
@specs.method
def dict_items(d):
    return six.iteritems(d)


@specs.parameter('lst', yaqltypes.Sequence(), alias='list')
@specs.parameter('index', int, nullable=False)
@specs.name('#indexer')
def list_indexer(lst, index):
    return lst[index]


@specs.parameter('value', nullable=True)
@specs.parameter('collection', yaqltypes.Iterable())
@specs.name('#operator_in')
def in_(value, collection):
    return value in collection


@specs.parameter('value', nullable=True)
@specs.parameter('collection', yaqltypes.Iterable())
@specs.method
def contains(collection, value):
    return value in collection


@specs.parameter('value', nullable=True)
@specs.parameter('d', utils.MappingType, alias='dict')
@specs.method
def contains_key(d, value):
    return value in d


@specs.parameter('value', nullable=True)
@specs.parameter('d', utils.MappingType, alias='dict')
@specs.method
def contains_value(d, value):
    return value in d.values()


@specs.parameter('left', yaqltypes.Iterable())
@specs.parameter('right', yaqltypes.Iterable())
@specs.name('#operator_+')
def combine_lists(left, right, engine):
    if isinstance(left, tuple) and isinstance(right, tuple):
        utils.limit_memory_usage(engine, (1, left), (1, right))
        return left + right

    elif isinstance(left, frozenset) and isinstance(right, frozenset):
        utils.limit_memory_usage(engine, (1, left), (1, right))
        return left.union(right)

    return yaql.standard_library.queries.concat(left, right)


@specs.parameter('left', yaqltypes.Sequence())
@specs.parameter('right', int)
@specs.name('#operator_*')
def list_by_int(left, right, engine):
    utils.limit_memory_usage(engine, (-right + 1, []), (right, left))
    return left * right


@specs.parameter('left', int)
@specs.parameter('right', yaqltypes.Sequence())
@specs.name('#operator_*')
def int_by_list(left, right, engine):
    return list_by_int(right, left, engine)


@specs.parameter('left', utils.MappingType)
@specs.parameter('right', utils.MappingType)
@specs.name('#operator_+')
def combine_dicts(left, right, engine):
    utils.limit_memory_usage(engine, (1, left), (1, right))
    d = dict(left)
    d.update(right)
    return utils.FrozenDict(d)


@specs.parameter('left', yaqltypes.Sequence())
@specs.parameter('right', yaqltypes.Sequence())
@specs.name('*equal')
def eq(left, right):
    return left == right


@specs.parameter('left', yaqltypes.Sequence())
@specs.parameter('right', yaqltypes.Sequence())
@specs.name('*not_equal')
def neq(left, right):
    return left != right


@specs.parameter('left', utils.MappingType)
@specs.parameter('right', utils.MappingType)
@specs.name('*equal')
def eq_dict(left, right):
    return left == right


@specs.parameter('left', utils.MappingType)
@specs.parameter('right', utils.MappingType)
@specs.name('*not_equal')
def neq_dict(left, right):
    return left != right


def is_list(arg):
    return utils.is_sequence(arg)


def is_dict(arg):
    return isinstance(arg, utils.MappingType)


def is_set(arg):
    return isinstance(arg, utils.SetType)


@specs.parameter('d', utils.MappingType, alias='dict')
@specs.extension_method
@specs.name('len')
def dict_len(d):
    return len(d)


@specs.parameter('sequence', yaqltypes.Sequence())
@specs.extension_method
@specs.name('len')
def sequence_len(sequence):
    return len(sequence)


@specs.method
@specs.parameter('collection', yaqltypes.Iterable())
@specs.parameter('position', int)
@specs.parameter('count', int)
def delete(collection, position, count=1):
    for i, t in enumerate(collection):
        if count >= 0 and not position <= i < position + count:
            yield t
        elif count < 0 and not i >= position:
            yield t


@specs.method
@specs.parameter('collection', yaqltypes.Iterable([
    lambda t: not is_set(t)
]))
@specs.parameter('position', int)
@specs.parameter('count', int)
def replace(collection, position, value, count=1):
    yielded = False
    for i, t in enumerate(collection):
        if (count >= 0 and position <= i < position + count
                or count < 0 and i >= position):
            if not yielded:
                yielded = True
                yield value
        else:
            yield t


@specs.method
@specs.parameter('collection', yaqltypes.Iterable())
@specs.parameter('position', int)
@specs.parameter('count', int)
@specs.parameter('values', yaqltypes.Iterable())
def replace_many(collection, position, values, count=1):
    yielded = False
    for i, t in enumerate(collection):
        if (count >= 0 and position <= i < position + count
                or count < 0 and i >= position):
            if not yielded:
                for v in values:
                    yield v
                yielded = True
        else:
            yield t


@specs.method
@specs.name('delete')
@specs.parameter('d', utils.MappingType, alias='dict')
def delete_keys(d, *keys):
    return delete_keys_seq(d, keys)


@specs.method
@specs.name('deleteAll')
@specs.parameter('d', utils.MappingType, alias='dict')
@specs.parameter('keys', yaqltypes.Iterable())
def delete_keys_seq(d, keys):
    copy = dict(d)
    for t in keys:
        copy.pop(t, None)
    return copy


@specs.method
@specs.parameter('collection', yaqltypes.Iterable(validators=[
    lambda x: not isinstance(x, utils.SetType)]
))
@specs.parameter('value', nullable=True)
@specs.parameter('position', int)
@specs.name('insert')
def iter_insert(collection, position, value):
    i = -1
    for i, t in enumerate(collection):
        if i == position:
            yield value
        yield t

    if position > i:
        yield value


@specs.method
@specs.parameter('collection', yaqltypes.Sequence())
@specs.parameter('value', nullable=True)
@specs.parameter('position', int)
@specs.name('insert')
def list_insert(collection, position, value):
    copy = list(collection)
    copy.insert(position, value)
    return copy


@specs.method
@specs.parameter('collection', yaqltypes.Iterable())
@specs.parameter('values', yaqltypes.Iterable())
@specs.parameter('position', int)
def insert_many(collection, position, values):
    i = -1
    if position < 0:
        for j in values:
            yield j
    for i, t in enumerate(collection):
        if i == position:
            for j in values:
                yield j
        yield t

    if position > i:
        for j in values:
            yield j


@specs.parameter('s', utils.SetType, alias='set')
@specs.extension_method
@specs.name('len')
def set_len(s):
    return len(s)


@specs.parameter('args', nullable=True)
@specs.inject('delegate', yaqltypes.Delegate('to_set', method=True))
def set_(delegate, *args):
    def rec(seq):
        for t in seq:
            if utils.is_iterator(t):
                for t2 in rec(t):
                    yield t2
            else:
                yield t
    return delegate(rec(args))


@specs.method
@specs.parameter('collection', yaqltypes.Iterable())
def to_set(collection):
    return frozenset(collection)


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.method
def union(left, right):
    return left.union(right)


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.name('*equal')
def set_eq(left, right):
    return left == right


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.name('*not_equal')
def set_neq(left, right):
    return left != right


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.name('#operator_<')
def set_lt(left, right):
    return left < right


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.name('#operator_<=')
def set_lte(left, right):
    return left <= right


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.name('#operator_>=')
def set_gte(left, right):
    return left >= right


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.name('#operator_>')
def set_gt(left, right):
    return left > right


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.method
def intersect(left, right):
    return left.intersection(right)


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.method
def difference(left, right):
    return left.difference(right)


@specs.parameter('left', utils.SetType)
@specs.parameter('right', utils.SetType)
@specs.method
def symmetric_difference(left, right):
    return left.symmetric_difference(right)


@specs.parameter('s', utils.SetType, alias='set')
@specs.method
@specs.name('add')
def set_add(s, *values):
    return s.union(frozenset(values))


@specs.parameter('s', utils.SetType, alias='set')
@specs.method
@specs.name('remove')
def set_remove(s, *values):
    return s.difference(frozenset(values))


def register(context, no_sets=False):
    context.register_function(list_)
    context.register_function(build_list)
    context.register_function(to_list)
    context.register_function(list_indexer)
    context.register_function(dict_)
    context.register_function(dict_, name='#map')
    context.register_function(dict__)
    context.register_function(to_dict)
    context.register_function(dict_keyword_access)
    context.register_function(dict_indexer)
    context.register_function(dict_indexer_with_default)
    context.register_function(dict_get)
    context.register_function(dict_keys)
    context.register_function(dict_values)
    context.register_function(dict_items)
    context.register_function(in_)
    context.register_function(contains_key)
    context.register_function(contains_value)
    context.register_function(combine_lists)
    context.register_function(collection_attribute)
    context.register_function(list_by_int)
    context.register_function(int_by_list)
    context.register_function(combine_dicts)
    context.register_function(eq)
    context.register_function(neq)
    context.register_function(eq_dict)
    context.register_function(neq_dict)
    context.register_function(is_dict)
    context.register_function(is_list)
    context.register_function(dict_len)
    context.register_function(sequence_len)
    context.register_function(delete)
    context.register_function(delete_keys)
    context.register_function(delete_keys_seq)
    context.register_function(iter_insert)
    context.register_function(list_insert)
    context.register_function(replace)
    context.register_function(replace_many)
    context.register_function(insert_many)
    context.register_function(contains)

    if not no_sets:
        context.register_function(is_set)
        context.register_function(set_)
        context.register_function(to_set)
        context.register_function(set_len)
        context.register_function(set_eq)
        context.register_function(set_neq)
        context.register_function(set_lt)
        context.register_function(set_lte)
        context.register_function(set_gt)
        context.register_function(set_gte)
        context.register_function(set_add)
        context.register_function(set_remove)
        context.register_function(union)
        context.register_function(intersect)
        context.register_function(difference)
        context.register_function(
            difference, name='#operator_-', function=True, method=False)
        context.register_function(symmetric_difference)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

import six

from yaql.language import expressions
from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes


@specs.parameter('left', yaqltypes.YaqlExpression())
@specs.name('#operator_=>')
def build_tuple(left, right, context, engine):
    if isinstance(left, expressions.BinaryOperator) and left.operator == '=>':
        return left(utils.NO_VALUE, context, engine)[0] + (right,)
    else:
        return left(utils.NO_VALUE, context, engine)[0], right


@specs.parameter('tuples', tuple)
@specs.inject('delegate', yaqltypes.Super(with_name=True))
@specs.no_kwargs
@specs.extension_method
def dict_(delegate, *tuples):
    return delegate('dict', tuples)


@specs.method
@specs.parameter('collection', yaqltypes.Iterable())
def to_list(collection):
    return list(collection)


def tuple_(*args):
    return args


@specs.parameter('collection', yaqltypes.Iterable())
@specs.parameter('index_expression', yaqltypes.Lambda())
def indexer(collection, index_expression):
    if isinstance(collection, utils.SequenceType):
        index = index_expression()
        if isinstance(index, int) and not isinstance(index, bool):
            return collection[index]
    return six.moves.filter(index_expression, collection)


@specs.parameter('start', int)
@specs.parameter('stop', int, nullable=True)
@specs.extension_method
def range_(start, stop=None):
    if stop is None:
        return itertools.count(start)
    else:
        return six.moves.range(start, stop)


@specs.parameter('conditions', yaqltypes.Lambda())
def switch(*conditions):
    for cond in conditions:
        res = cond()
        if not isinstance(res, tuple):
            raise ValueError('switch() must have tuple parameters')
        if len(res) != 2:
            raise ValueError('switch() tuples must be of size 2')
        if res[0]:
            return res[1]
    return None


@specs.parameter('mappings', yaqltypes.Lambda())
@specs.method
def as_(context, sender, *mappings):
    for t in mappings:
        tt = t(sender)
        if not isinstance(tt, tuple):
            raise ValueError('as() must have tuple parameters')
        if len(tt) != 2:
            raise ValueError('as() tuples must be of size 2')
        context[tt[1]] = tt[0]
    return sender


def _to_extension_method(name, context):
    for spec in context.parent.get_functions(
            name, lambda t: not t.is_function or not t.is_method,
            use_convention=True):
        spec = spec.clone()
        spec.is_function = True
        spec.is_method = True
        context.register_function(spec)


def register(context, tuples):
    if tuples:
        context.register_function(build_tuple)
        context.register_function(to_list)
        context.register_function(tuple_)

    context.register_function(dict_)
    context.register_function(dict_, name='#map')
    context.register_function(indexer, name='#indexer', exclusive=True)
    context.register_function(range_)
    context.register_function(switch, exclusive=True)
    context.register_function(as_)

    for t in ('get', 'list', 'bool', 'int', 'float', 'select', 'where',
              'join', 'sum', 'take_while'):
        _to_extension_method(t, context)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaql.tests


class TestLegacy(yaql.tests.TestCase):
    def setUp(self):
        super(TestLegacy, self).setUp()
        self.eval = self.legacy_eval

    def test_tuples(self):
        self.assertEqual((1, 2), self.eval('1 => 2'))
        self.assertEqual((None, 'a b'), self.eval('null => "a b"'))
        self.assertEqual((1, 2, 3), self.eval('1 => 2 => 3'))
        self.assertEqual(((1, 2), 3), self.eval('(1 => 2) => 3'))
        self.assertEqual((1, (2, 3)), self.eval('1 => (2 => 3)'))

    def test_tuples_func(self):
        self.assertEqual((1, 2), self.eval('tuple(1, 2)'))
        self.assertEqual((None,), self.eval('tuple(null)'))
        self.assertEqual(tuple(), self.eval('tuple()'))

    def test_dict(self):
        self.assertEqual(
            {'a': 'b', 1: 2, None: None},
            self.eval('dict(1 => 2, a => b, null => null)'))

        self.assertEqual({}, self.eval('dict()'))

    def test_list(self):
        self.assertEqual([1, 2, 'a', None], self.eval('list(1, 2, a, null)'))
        self.assertEqual([], self.eval('list()'))
        self.assertEqual([1, 2], self.eval('[1, 2].select($).list()'))

    def test_dict_get(self):
        self.assertEqual(5, self.eval("get($, 'a b')", data={'a b': 5}))
        self.assertEqual(5, self.eval("$.get('a b')", data={'a b': 5}))

    def test_int(self):
        self.assertEqual(5, self.eval("'5'.int()"))
        self.assertEqual(5, self.eval("5.2.int()"))
        self.assertEqual(5, self.eval("int('5')"))
        self.assertEqual(5, self.eval("int(5.2)"))

    def test_float(self):
        self.assertAlmostEqual(5.1, self.eval("'5.1'.float()"))
        self.assertAlmostEqual(5.0, self.eval("5.float()"))
        self.assertAlmostEqual(5.1, self.eval("float('5.1')"))
        self.assertAlmostEqual(5.0, self.eval("float(5)"))

    def test_bool(self):
        self.assertFalse(self.eval("null.bool()"))
        self.assertFalse(self.eval("''.bool()"))
        self.assertFalse(self.eval("0.bool()"))
        self.assertFalse(self.eval("false.bool()"))
        self.assertFalse(self.eval("[].bool()"))
        self.assertFalse(self.eval("{}.bool()"))
        self.assertTrue(self.eval("' '.bool()"))
        self.assertTrue(self.eval("x.bool()"))
        self.assertTrue(self.eval("1.bool()"))
        self.assertTrue(self.eval("true.bool()"))
        self.assertTrue(self.eval("[1].bool()"))
        self.assertTrue(self.eval("{a=>b}.bool()"))

    def test_filter(self):
        self.assertEqual(2, self.eval("list(1,2,3)[1]"))
        self.assertEqual(3, self.eval("list(1,2,3)[$]", data=2))
        self.assertEqual([1, 3], self.eval("list(1,2,3)[$ != 2]"))
        self.assertEqual([], self.eval("list()[$ > 0]"))

    def test_sum(self):
        self.assertEqual(6, self.eval('list(1,2,3).sum()'))
        self.assertEqual(6, self.eval('sum(list(1,2,3))'))

    def test_range(self):
        self.assertEqual([2, 3, 4, 5], self.eval('range(2).take(4)'))
        self.assertEqual([1, 2, 3], self.eval('range(1, 4)'))
        self.assertEqual([2, 3, 4, 5], self.eval('2.range().take(4)'))
        self.assertEqual([1, 2, 3], self.eval('1.range(4)'))

    def test_take_while(self):
        self.assertEqual([1, 2], self.eval('[1, 2, 3, 4].takeWhile($ < 3)'))
        self.assertEqual([1, 2], self.eval('takeWhile([1, 2, 3, 4], $ < 3)'))

    def test_switch(self):
        expr = 'switch($ < 10 => 1, $ >= 10 and $ < 100 => 2, $ >= 100 => 3)'
        self.assertEqual(3, self.eval(expr, data=123))
        self.assertEqual(2, self.eval(expr, data=50))
        self.assertEqual(1, self.eval(expr, data=-123))

    def test_as(self):
        self.assertEqual(
            [3, 6],
            self.eval('[1, 2].as(sum($) => a).select($ * $a)'))



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import six

from yaql.language import exceptions
from yaql.language import expressions
from yaql.language import utils


class HiddenParameterType(object):
    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def check(self, value):
        return True


class LazyParameterType(object):
    pass


class SmartType(object):
    def __init__(self, nullable):
        self.nullable = nullable

    def check(self, value):
        if value is None and not self.nullable:
            return False
        return True

    def convert(self, value, sender, context, function_spec, engine):
        if not self.check(value):
            raise exceptions.ArgumentValueException()
        utils.limit_memory_usage(engine, (1, value))

    def is_specialization_of(self, other):
        return False


class PythonType(SmartType):
    def __init__(self, python_type, nullable=True, validators=None):
        self.python_type = python_type
        if not validators:
            validators = [lambda _: True]
        if not isinstance(validators, (list, tuple)):
            validators = [validators]
        self.validators = validators
        super(PythonType, self).__init__(nullable)

    def check(self, value):
        if isinstance(value, expressions.Constant):
            value = value.value

        return super(PythonType, self).check(value) and (
            value is None or isinstance(value, expressions.Expression) or (
                isinstance(value, self.python_type) and all(
                    map(lambda t: t(value), self.validators))))

    def convert(self, value, sender, context, function_spec, engine):
        super(PythonType, self).convert(value, sender, context,
                                        function_spec, engine)
        if isinstance(value, expressions.Constant):
            value = value.value
        super(PythonType, self).convert(value, sender, context,
                                        function_spec, engine)
        return value

    def is_specialization_of(self, other):
        return (
            isinstance(other, PythonType)
            and issubclass(self.python_type, other.python_type)
            and not issubclass(other.python_type, self.python_type)
        )


class MappingRule(LazyParameterType, SmartType):
    def __init__(self):
        super(MappingRule, self).__init__(False)

    def check(self, value):
        return isinstance(value, expressions.MappingRuleExpression)

    def convert(self, value, sender, context, function_spec, engine):
        super(MappingRule, self).convert(value, sender, context,
                                         function_spec, engine)
        wrap = lambda func: lambda: func(sender, context, engine)[0]

        return utils.MappingRule(wrap(value.source), wrap(value.destination))


class String(PythonType):
    def __init__(self, nullable=False):
        super(String, self).__init__(six.string_types, nullable=nullable)

    def convert(self, value, sender, context, function_spec, engine):
        value = super(String, self).convert(value, sender, context,
                                            function_spec, engine)
        return None if value is None else six.text_type(value)


class Iterable(PythonType):
    def __init__(self, validators=None):
        super(Iterable, self).__init__(
            collections.Iterable, False, [
                lambda t: not isinstance(t, six.string_types + (
                    utils.MappingType,))] + (validators or []))

    def convert(self, value, sender, context, function_spec, engine):
        res = super(Iterable, self).convert(value, sender, context,
                                            function_spec, engine)
        return utils.limit_iterable(res, engine)


class Iterator(Iterable):
    def __init__(self, validators=None):
        super(Iterator, self).__init__(
            validators=[utils.is_iterator] + (validators or []))


class Sequence(PythonType):
    def __init__(self, validators=None):
        super(Sequence, self).__init__(
            collections.Sequence, False, [
                lambda t: not isinstance(t, six.string_types + (dict,))] + (
                    validators or []))


class Number(PythonType):
    def __init__(self, nullable=False):
        super(Number, self).__init__(
            six.integer_types + (float,), nullable, [
                lambda t: type(t) is not bool])


class Lambda(LazyParameterType, SmartType):
    def __init__(self, with_context=False, method=False, return_context=False):
        super(Lambda, self).__init__(True)
        self.return_context = return_context
        self.with_context = with_context
        self.method = method

    def check(self, value):
        if self.method and isinstance(
                value, expressions.Expression) and not value.uses_sender:
            return False
        return super(Lambda, self).check(value)

    @staticmethod
    def _publish_params(context, args, kwargs):
        for i, param in enumerate(args):
            context['$' + str(i + 1)] = param
        for arg_name, arg_value in kwargs.items():
            context['$' + arg_name] = arg_value

    def _call(self, value, sender, context, engine, args, kwargs):
        self._publish_params(context, args, kwargs)
        if isinstance(value, expressions.Expression):
            result = value(sender, context, engine)
        elif six.callable(value):
            result = value, context
        else:
            result = value, context
        return result[0] if not self.return_context else result

    def convert(self, value, sender, context, function_spec, engine):
        super(Lambda, self).convert(value, sender, context,
                                    function_spec, engine)
        if value is None:
            return None

        def func(*args, **kwargs):
            if self.method and self.with_context:
                new_sender, new_context = args[:2]
                args = args[2:]
            elif self.method and not self.with_context:
                new_sender, new_context = \
                    args[0], context.create_child_context()
                args = args[1:]
            elif not self.method and self.with_context:
                new_sender, new_context = utils.NO_VALUE, args[0]
                args = args[1:]
            else:
                new_sender, new_context = \
                    utils.NO_VALUE, context.create_child_context()

            return self._call(value, new_sender, new_context,
                              engine, args, kwargs)

        return func


class Super(HiddenParameterType, SmartType):
    def __init__(self, with_context=False, method=None, return_context=False,
                 with_name=False):
        self.return_context = return_context
        self.with_context = with_context
        self.method = method
        self.with_name = with_name
        super(Super, self).__init__(False)

    @staticmethod
    def _find_function_context(spec, context):
        while context is not None:
            funcs = context.get_functions(spec.name)
            if funcs and spec in funcs:
                return context
            context = context.parent
        raise exceptions.NoFunctionRegisteredException(
            spec.name)

    def convert(self, value, sender, context, function_spec, engine):
        def func(*args, **kwargs):
            function_context = self._find_function_context(
                function_spec, context)
            parent_function_context = function_context.parent
            if parent_function_context is None:
                raise exceptions.NoFunctionRegisteredException(
                    function_spec.name)

            new_name = function_spec.name
            if self.with_name:
                new_name = args[0]
                args = args[1:]

            new_sender = sender
            if self.method is True:
                new_sender = args[0]
                args = args[1:]
            elif self.method is False:
                new_sender = utils.NO_VALUE

            if self.with_context:
                new_context = args[0]
                args = args[1:]
            else:
                new_context = context.create_child_context()

            return parent_function_context(
                new_name, engine, new_sender, new_context,
                self.return_context)(*args, **kwargs)
        return func


class Context(HiddenParameterType, SmartType):
    def __init__(self):
        super(Context, self).__init__(False)

    def convert(self, value, sender, context, function_spec, engine):
        return context


class Delegate(HiddenParameterType, SmartType):
    def __init__(self, name=None, with_context=False, method=False,
                 return_context=False):
        super(Delegate, self).__init__(False)
        self.name = name
        self.return_context = return_context
        self.with_context = with_context
        self.method = method

    def convert(self, value, sender, context, function_spec, engine):
        def func(*args, **kwargs):
            name = self.name
            if not name:
                name = args[0]
                args = args[1:]

            new_sender = utils.NO_VALUE
            if self.method:
                new_sender = args[0]
                args = args[1:]
            if self.with_context:
                new_context = args[0]
                args = args[1:]
            else:
                new_context = context.create_child_context()

            return new_context(
                name, engine, new_sender, return_context=self.return_context,
                use_convention=True)(*args, **kwargs)
        return func


class Sender(HiddenParameterType, SmartType):
    def __init__(self):
        super(Sender, self).__init__(False)

    def convert(self, value, sender, context, function_spec, engine):
        return sender


class Engine(HiddenParameterType, SmartType):
    def __init__(self):
        super(Engine, self).__init__(False)

    def convert(self, value, sender, context, function_spec, engine):
        return engine


class FunctionDefinition(HiddenParameterType, SmartType):
    def __init__(self):
        super(FunctionDefinition, self).__init__(False)

    def convert(self, value, sender, context, function_spec, engine):
        return function_spec


class Constant(SmartType):
    def __init__(self, nullable, expand=True):
        self.expand = expand
        super(Constant, self).__init__(nullable)

    def check(self, value):
        return super(Constant, self).check(value.value) and (
            value is None or isinstance(value, expressions.Constant))

    def convert(self, value, sender, context, function_spec, engine):
        super(Constant, self).convert(value, sender, context,
                                      function_spec, engine)
        return value.value if self.expand else value


class YaqlExpression(LazyParameterType, SmartType):
    def __init__(self):
        super(YaqlExpression, self).__init__(False)

    def check(self, value):
        return isinstance(value, expressions.Expression)

    def convert(self, value, sender, context, function_spec, engine):
        super(YaqlExpression, self).convert(value, sender, context,
                                            function_spec, engine)
        return value


class StringConstant(Constant):
    def __init__(self, nullable=False):
        super(StringConstant, self).__init__(nullable)

    def check(self, value):
        return super(StringConstant, self).check(value) and (
            value is None or isinstance(value.value, six.string_types))


class Keyword(Constant):
    def __init__(self, expand=True):
        super(Keyword, self).__init__(False, expand)

    def check(self, value):
        return isinstance(value, expressions.KeywordConstant)


class BooleanConstant(Constant):
    def __init__(self, nullable=False, expand=True):
        super(BooleanConstant, self).__init__(nullable, expand)

    def check(self, value):
        return super(BooleanConstant, self).check(value) and (
            value is None or type(value.value) is bool)


class NumericConstant(Constant):
    def __init__(self, nullable=False, expand=True):
        super(NumericConstant, self).__init__(nullable, expand)

    def check(self, value):
        return super(NumericConstant, self).check(value) and (
            value is None or isinstance(
                value.value, six.integer_types + (float,)) and
            type(value.value) is not bool)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import exceptions
from yaql.language import specs
from yaql.language import yaqltypes
import yaql.tests


class TestResolution(yaql.tests.TestCase):
    def test_resolve_parameter_count_single_layer(self):
        def f1(a):
            return a

        def f2(a, b):
            return a + b

        self.context.register_function(f1, name='f')
        self.context.register_function(f2, name='f')

        self.assertEqual(12, self.eval('f(12)'))
        self.assertEqual(25, self.eval('f(12, 13)'))

    def test_resolve_parameter_count_multi_layer(self):
        def f1(a):
            return a

        def f2(a, b):
            return a + b

        context1 = self.context.create_child_context()
        context1.register_function(f1, name='f')
        context2 = context1.create_child_context()
        context2.register_function(f2, name='f')

        self.assertEqual(12, self.eval('f(12)', context=context2))
        self.assertEqual(25, self.eval('f(12, 13)', context=context2))

    def test_layer_override(self):
        def f1(a):
            return a

        def f2(a):
            return -a

        context1 = self.context.create_child_context()
        context1.register_function(f1, name='f')
        context2 = context1.create_child_context()
        context2.register_function(f2, name='f')

        self.assertEqual(-12, self.eval('f(12)', context=context2))

    def test_single_layer_ambiguity(self):
        def f1(a):
            return a

        def f2(a):
            return -a

        context1 = self.context.create_child_context()
        context1.register_function(f1, name='f')
        context1.register_function(f2, name='f')

        self.assertRaises(
            exceptions.AmbiguousFunctionException,
            self.eval, 'f(12)', context=context1)

    def test_single_layer_laziness_ambiguity(self):
        @specs.parameter('a', yaqltypes.Lambda())
        def f1(a):
            return a()

        def f2(a):
            return -a

        def f3(a, b):
            return a + b

        context1 = self.context.create_child_context()
        context1.register_function(f1, name='f')
        context1.register_function(f2, name='f')
        context1.register_function(f3, name='f')

        self.assertRaises(
            exceptions.AmbiguousFunctionException,
            self.eval, 'f(2 * $)', data=3, context=context1)

        self.assertEqual(25, self.eval('f(12, 13)', context=context1))

    def test_multi_layer_laziness_ambiguity(self):
        @specs.parameter('a', yaqltypes.Lambda())
        def f1(a, b):
            return a() + b

        @specs.parameter('a', yaqltypes.Lambda())
        def f2(a, b):
            return a() + b

        @specs.parameter('b', yaqltypes.Lambda())
        def f3(a, b):
            return -a - b()

        @specs.parameter('a', yaqltypes.Lambda())
        def f4(a, b):
            return -a() + b

        context1 = self.context.create_child_context()
        context1.register_function(f1, name='foo')
        context1.register_function(f2, name='bar')
        context2 = context1.create_child_context()
        context2.register_function(f3, name='foo')
        context2.register_function(f4, name='bar')

        self.assertRaises(
            exceptions.AmbiguousFunctionException,
            self.eval, 'foo(12, 13)', context=context2)

        self.assertEqual(
            1,
            self.eval('bar(12, 13)', context=context2))

    def test_ambiguous_method(self):
        self.context.register_function(
            lambda c, s: 1, name='select', method=True)
        self.assertRaises(
            exceptions.AmbiguousMethodException,
            self.eval, '[1,2].select($)')



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class YaqlException(Exception):
    pass


class ResolutionError(YaqlException):
    pass


class FunctionResolutionError(ResolutionError):
    pass


class MethodResolutionError(ResolutionError):
    pass


class NoFunctionRegisteredException(FunctionResolutionError):
    def __init__(self, name):
        super(NoFunctionRegisteredException, self).__init__(
            u'Unknown function "{0}"'.format(name))


class NoMethodRegisteredException(MethodResolutionError):
    def __init__(self, name, sender):
        super(NoMethodRegisteredException, self).__init__(
            u'Unknown method "{0}" for type {1}'.format(name, type(sender)))


class NoMatchingFunctionException(FunctionResolutionError):
    def __init__(self, name):
        super(NoMatchingFunctionException, self).__init__(
            u'No function "{0}" matches supplied arguments'.format(name))


class NoMatchingMethodException(MethodResolutionError):
    def __init__(self, name, sender):
        super(NoMatchingMethodException, self).__init__(
            u'No method "{0}" for type {1} matches supplied arguments'.format(
                name, type(sender)))


class AmbiguousFunctionException(FunctionResolutionError):
    def __init__(self, name):
        super(AmbiguousFunctionException, self).__init__(
            u'Ambiguous function "{0}"'.format(name))


class AmbiguousMethodException(MethodResolutionError):
    def __init__(self, name, sender):
        super(AmbiguousMethodException, self).__init__(
            u'Ambiguous method "{0}" for type {1}'.format(name, type(sender)))


class ArgumentException(YaqlException):
    def __init__(self, argument_name):
        self.parameter_name = argument_name
        super(ArgumentException, self).__init__(
            u'Invalid argument {0}'.format(argument_name))


class MappingTranslationException(YaqlException):
    def __init__(self):
        super(MappingTranslationException, self).__init__(
            u'Cannot convert mapping to keyword argument')


class ArgumentValueException(YaqlException):
    def __init__(self):
        super(ArgumentValueException, self).__init__()


class DuplicateParameterDecoratorException(YaqlException):
    def __init__(self, function_name, param_name):
        message = u"Function '{0}' has multiple " \
                  u"decorators for parameter '{1}'". \
            format(function_name, param_name)
        super(DuplicateParameterDecoratorException, self).__init__(message)


class InvalidMethodException(YaqlException):
    def __init__(self, function_name):
        message = u"Function '{0}' cannot be called as a method". \
            format(function_name)
        super(InvalidMethodException, self).__init__(message)


class NoParameterFoundException(YaqlException):
    def __init__(self, function_name, param_name):
        message = u"Function '{0}' has no parameter called '{1}'". \
            format(function_name, param_name)
        super(NoParameterFoundException, self).__init__(message)


class YaqlParsingException(YaqlException):
    def __init__(self, value, position, message):
        self.value = value
        self.position = position
        self.message = message
        super(YaqlParsingException, self).__init__(message)


class YaqlGrammarException(YaqlParsingException):
    def __init__(self, expr, value, position):
        if position is None:
            msg = u'Parse error: unexpected end of statement'
        else:
            msg = u"Parse error: unexpected '{0}' at position {1} of " \
                  u"expression '{2}'".format(value, position, expr)
        super(YaqlGrammarException, self).__init__(value, position, msg)


class YaqlLexicalException(YaqlParsingException):
    def __init__(self, value, position):
        msg = u"Lexical error: illegal character '{0}' at position {1}" \
            .format(value, position)
        super(YaqlLexicalException, self).__init__(value, position, msg)


class InvalidOperatorTableException(YaqlException):
    def __init__(self, op):
        super(InvalidOperatorTableException, self). \
            __init__(u"Invalid records in operator table for operator "
                     u"'{0}".format(op))


class WrappedException(YaqlException):
    def __init__(self, exception):
        self.wrapped = exception
        super(WrappedException, self).__init__(str(exception))


class CollectionTooLargeException(YaqlException):
    def __init__(self, count):
        super(CollectionTooLargeException, self).__init__(
            'Collection length exceeds {0} elements'.format(count))


class MemoryQuotaExceededException(YaqlException):
    def __init__(self):
        super(MemoryQuotaExceededException, self).__init__(
            'Expression consumed too much memory')



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaql.tests


class TestSystem(yaql.tests.TestCase):
    def test_def(self):
        self.assertEqual(
            [1, 4, 9],
            self.eval('def(sq, $*$) -> $.select(sq($))', data=[1, 2, 3]))
        self.assertEqual(
            [1, 4, 9],
            self.eval('def(sq, $arg * $arg) -> $.select(sq(arg => $))',
                      data=[1, 2, 3]))

    def test_def_recursion(self):
        self.assertEqual(24, self.eval(
            'def(rec, switch($ = 1 => 1, true => $*rec($-1))) -> rec($)',
            data=4))

    def test_elvis_dict(self):
        self.assertEqual(1, self.eval('$?.a', data={'a': 1}))
        self.assertIsNone(self.eval('$?.a', data=None))

    def test_elvis_method(self):
        self.assertEqual([2, 3], self.eval('$?.select($+1)', data=[1, 2]))
        self.assertIsNone(self.eval('$?.select($+1)', data=None))

    def test_unpack(self):
        self.assertEqual(
            5, self.eval('[2, 3].unpack() -> $1 + $2'))

    def test_unpack_with_names(self):
        self.assertEqual(
            5, self.eval('[2, 3].unpack(a, b) -> $a + $b'))

        self.assertRaises(
            ValueError,
            self.eval, '[2, 3].unpack(a, b, c) -> $a + $b')

        self.assertRaises(
            ValueError,
            self.eval, '[2, 3].unpack(a) -> $a')

    def test_assert(self):
        self.assertEqual(
            [3, 4],
            self.eval('[2, 3].assert(len($) > 1).select($ + 1)'))

        self.assertRaises(
            AssertionError,
            self.eval, '[2].assert(len($) > 1).select($ + 1)')

        self.assertEqual(
            3,
            self.eval('[2].select($ + 1).assert(len($) = 1).first()'))
